==> pftpd-0.37.tar.gz: pftpd-0.37/src/rpad.c <==
/*
** rpad.c - Reserved Port Allocation service Daemon
**
** Copyright (c) 2002 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#ifdef HAVE_DOOR_CALL

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>

#ifdef HAVE_DOOR_H
#include <door.h>
#endif

#include "rpa.h"

#include "plib/safeio.h"

#define INIT_PID 1

int debug = 0;
int init_mode = 0;

int access_euid = -1;
int access_egid = -1;
int access_ruid = -1;
int access_rgid = -1;

#define PORT_FTP_DATA 20


int
create_bind_socket(struct sockaddr_gen *sin)
{
    int fd, try;
    static int one = 1;
    

    fd = socket(SGFAM(*sin), SOCK_STREAM, 0);
    if (fd < 0)
    {
	syslog(LOG_ERR, "socket() failed: %m");
	return -1;
    }
    
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *) &one, sizeof(one)))
    {
	syslog(LOG_WARNING, "setsockopt(SO_REUSEADDR) failed (ignored): %m");
    }
    
    SGPORT(*sin) = PORT_FTP_DATA;
    
    try = 0;
    while (s_bind(fd, (struct sockaddr *) sin, SGSOCKSIZE(*sin)) < 0)
    {
	char buf1[256];
	
	if (debug)
	{
	    fprintf(stderr,
		    "ftpdata_run: s_bind(): try=%d, addr=%s, failed: %s\n",
		    try,
		    s_inet_ntox(sin, buf1, sizeof(buf1)),
		    strerror(errno));
	}
	
	if (try++ > 10 || errno != EADDRINUSE)
	{
	    char buf1[256];
	    
	    
	    syslog(LOG_WARNING, "bind(%d,%s): %m",
		   fd,
		   s_inet_ntox(sin, buf1, sizeof(buf1)));
	    
	    s_close(fd);
	    return -1;
	}
	
	sleep(try);
    }
    
    return fd;
}


void
servproc(void *cookie,
	 char *dataptr,
	 size_t datasize,
	 door_desc_t *descptr,
	 size_t ndesc)
{
    door_desc_t db;
    door_cred_t cb;
    struct sockaddr_gen sin;
    int rcode, rfd;



    if (debug)
	fprintf(stderr, "rpad: servproc: Start\n");
    
    door_cred(&cb);

    if (access_euid != -1 && access_euid != cb.dc_euid)
	goto Fail;
    
    if (access_egid != -1 && access_egid != cb.dc_egid)
	goto Fail;
    
    if (access_ruid != -1 && access_ruid != cb.dc_ruid)
	goto Fail;
    
    if (access_rgid != -1 && access_rgid != cb.dc_rgid)
	goto Fail;
    
    if (dataptr == DOOR_UNREF_DATA)
    {
	if (debug)
	    printf("rpad: servproc: door unreferenced\n");
	
	goto Fail;
    }

    sin = * (struct sockaddr_gen *) dataptr;
    
    rfd = create_bind_socket(&sin);
    rcode = errno;
    
    db.d_data.d_desc.d_descriptor = rfd;
    db.d_attributes = DOOR_DESCRIPTOR|DOOR_RELEASE;
    
    if (debug)
	printf("rpad: servproc: end\n");
    
    door_return((char *) &rcode,
		sizeof(rcode),
		&db, 1);

  Fail:
    door_return(NULL, 0, NULL, 0);
}


int
main(int argc,
     char *argv[])
{
    int dfd, pfd, i, j;
    char path_door[1024];
    char *service = "ftp";
    

    for (i = 1; i < argc && argv[i][0] == '-'; ++i)
	for (j = 1; j > 0 && argv[i][j]; ++j)
	    switch (argv[i][j])
	    {
	      case '-':
		goto EndOptions;
		
	      case 'd':
		++debug;
		break;

	      case 's':
		if (argv[i][j+1])
		{
		    service = strdup(argv[i]+j+1);
		    j = -1;
		}
		else if (i+1 < argc)
		{
		    service = strdup(argv[++i]);
		}
		else
		{
		    fprintf(stderr, "%s: -%c: missing service name\n",
			    argv[0], argv[i][j]);
		    exit(1);
		}
		break;
		
	      case 'I':
		++init_mode;
		break;
		
	      default:
		fprintf(stderr, "%s: unknown command line option: -%c\n",
			argv[0], argv[i][j]);
		exit(1);
	    }

  EndOptions:

    if (i < argc)
    {
	fprintf(stderr, "usage: %0 [<options>]\n", argv[0]);
	exit(1);
    }

    sprintf(path_door, "%s/%s/door", PATH_RPAD_DIR, service);
    if (unlink(path_door) < 0 && errno != ENOENT)
    {
	fprintf(stderr, "%s: unlink(\"%s\") failed: %s\n",
		argv[0], path_door, strerror(errno));
	exit(1);
    }
    
    pfd = open(path_door, O_CREAT|O_RDWR, 0644);
    if (pfd < 0)
    {
	fprintf(stderr, "%s: open(\"%s\", O_CREAT|O_RDWR) failed: %s\n",
		argv[0], path_door, strerror(errno));
	exit(1);
    }
    (void) close(pfd);
    
    if (!debug && 
	getppid() != INIT_PID && !init_mode)
    {
	become_daemon();
    }

    dfd = door_create(servproc, NULL, DOOR_UNREF);
    if (dfd < 0)
    {
	syslog(LOG_ERR, "door_create() failed: %s\n", strerror(errno));
	(void) unlink(path_door);
	exit(1);
    }
    
    if (fattach(dfd, path_door) < 0)
    {
	syslog(LOG_ERR, "fattach(\"%s\") failed: %s\n",
	       path_door, strerror(errno)); 

	(void) unlink(path_door);
	exit(1);
    }
    
    if (debug)
	fprintf(stderr, "Waiting for DOOR calls...\n");
    
    while (1)
	pause();

    syslog(LOG_CRIT, "forever loop terminated - exiting");
    
    fprintf(stderr, "%s: Internal Error: Should not happen!\n", argv[0]);
    exit(1);
}

#else

#error RPAD needs the DOORS IPC mechanism to compile and build.

#endif

==> pftpd-0.39.tar.gz: pftpd-0.39/src/rpad.c <==
/*
** rpad.c - Reserved Port Allocation service Daemon
**
** Copyright (c) 2002 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#ifdef HAVE_DOOR_CALL

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>

#ifdef HAVE_DOOR_H
#include <door.h>
#endif

#include "rpa.h"

#include "plib/safeio.h"
#include "plib/daemon.h"

#define INIT_PID 1

int debug = 0;
int init_mode = 0;

int access_euid = -1;
int access_egid = -1;
int access_ruid = -1;
int access_rgid = -1;

#define PORT_FTP_DATA 20


int
create_bind_socket(struct sockaddr_gen *sin)
{
    int fd, try;
    static int one = 1;
    

    fd = socket(SGFAM(*sin), SOCK_STREAM, 0);
    if (fd < 0)
    {
	syslog(LOG_ERR, "socket() failed: %m");
	return -1;
    }
    
    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *) &one, sizeof(one)))
    {
	syslog(LOG_WARNING, "setsockopt(SO_REUSEADDR) failed (ignored): %m");
    }
    
    SGPORT(*sin) = PORT_FTP_DATA;
    
    try = 0;
    while (s_bind(fd, (struct sockaddr *) sin, SGSOCKSIZE(*sin)) < 0)
    {
	char buf1[256];
	
	if (debug)
	{
	    fprintf(stderr,
		    "ftpdata_run: s_bind(): try=%d, addr=%s, failed: %s\n",
		    try,
		    s_inet_ntox(sin, buf1, sizeof(buf1)),
		    strerror(errno));
	}
	
	if (try++ > 10 || errno != EADDRINUSE)
	{
	    char buf1[256];
	    
	    
	    syslog(LOG_WARNING, "bind(%d,%s): %m",
		   fd,
		   s_inet_ntox(sin, buf1, sizeof(buf1)));
	    
	    s_close(fd);
	    return -1;
	}
	
	sleep(try);
    }
    
    return fd;
}


void
servproc(void *cookie,
	 char *dataptr,
	 size_t datasize,
	 door_desc_t *descptr,
	 size_t ndesc)
{
    door_desc_t db;
    door_cred_t cb;
    struct sockaddr_gen sin;
    int rcode, rfd;



    if (debug)
	fprintf(stderr, "rpad: servproc: Start\n");
    
    door_cred(&cb);

    if (access_euid != -1 && access_euid != cb.dc_euid)
	goto Fail;
    
    if (access_egid != -1 && access_egid != cb.dc_egid)
	goto Fail;
    
    if (access_ruid != -1 && access_ruid != cb.dc_ruid)
	goto Fail;
    
    if (access_rgid != -1 && access_rgid != cb.dc_rgid)
	goto Fail;
    
    if (dataptr == DOOR_UNREF_DATA)
    {
	if (debug)
	    printf("rpad: servproc: door unreferenced\n");
	
	goto Fail;
    }

    sin = * (struct sockaddr_gen *) dataptr;
    
    rfd = create_bind_socket(&sin);
    
    if (rfd < 0)
	rcode = errno;
    else
    {
	db.d_data.d_desc.d_descriptor = rfd;
	db.d_attributes = DOOR_DESCRIPTOR|DOOR_RELEASE;
    }
	
    if (debug)
	printf("rpad: servproc: end\n");
    
    door_return((char *) &rcode,
		sizeof(rcode),
		&db, 1);

  Fail:
    door_return(NULL, 0, NULL, 0);
}


int
main(int argc,
     char *argv[])
{
    int dfd, pfd, i, j;
    char path_door[1024];
    char *service = "ftp";
    

    for (i = 1; i < argc && argv[i][0] == '-'; ++i)
	for (j = 1; j > 0 && argv[i][j]; ++j)
	    switch (argv[i][j])
	    {
	      case '-':
		goto EndOptions;
		
	      case 'd':
		++debug;
		break;

	      case 's':
		if (argv[i][j+1])
		{
		    service = strdup(argv[i]+j+1);
		    j = -1;
		}
		else if (i+1 < argc)
		{
		    service = strdup(argv[++i]);
		}
		else
		{
		    fprintf(stderr, "%s: -%c: missing service name\n",
			    argv[0], argv[i][j]);
		    exit(1);
		}
		break;
		
	      case 'I':
		++init_mode;
		break;
		
	      default:
		fprintf(stderr, "%s: unknown command line option: -%c\n",
			argv[0], argv[i][j]);
		exit(1);
	    }

  EndOptions:

    if (i < argc)
    {
	fprintf(stderr, "usage: %s [<options>]\n", argv[0]);
	exit(1);
    }

    sprintf(path_door, "%s/%s/door", PATH_RPAD_DIR, service);
    if (unlink(path_door) < 0 && errno != ENOENT)
    {
	fprintf(stderr, "%s: unlink(\"%s\") failed: %s\n",
		argv[0], path_door, strerror(errno));
	exit(1);
    }
    
    pfd = open(path_door, O_CREAT|O_RDWR, 0644);
    if (pfd < 0)
    {
	fprintf(stderr, "%s: open(\"%s\", O_CREAT|O_RDWR) failed: %s\n",
		argv[0], path_door, strerror(errno));
	exit(1);
    }
    (void) close(pfd);
    
    if (!debug && 
	getppid() != INIT_PID && !init_mode)
    {
	become_daemon();
    }

    dfd = door_create(servproc, NULL, DOOR_UNREF);
    if (dfd < 0)
    {
	syslog(LOG_ERR, "door_create() failed: %s\n", strerror(errno));
	(void) unlink(path_door);
	exit(1);
    }
    
    if (fattach(dfd, path_door) < 0)
    {
	syslog(LOG_ERR, "fattach(\"%s\") failed: %s\n",
	       path_door, strerror(errno)); 

	(void) unlink(path_door);
	exit(1);
    }
    
    if (debug)
	fprintf(stderr, "Waiting for DOOR calls...\n");
    
    while (1)
	pause();

    syslog(LOG_CRIT, "forever loop terminated - exiting");
    
    fprintf(stderr, "%s: Internal Error: Should not happen!\n", argv[0]);
    exit(1);
}

#else

#error RPAD needs the DOORS IPC mechanism to compile and build.

#endif
